class TestConstants:
    YOLOV5S_MODEL_PATH = "yolov5/weights/yolov5s.pt"
    YOLOV5L_MODEL_PATH = "yolov5/weights/yolov5l.pt"
    ZIDANE_IMAGE_PATH = "tests/data/zidane.jpg"
    BUS_IMAGE_PATH = "tests/data/bus.jpg"
